/*** Z80Em: Portable Z80 emulator *******************************************/
/***                                                                      ***/
/***                                 Z80.c                                ***/
/***                                                                      ***/
/*** This file contains the emulation code                                ***/
/***                                                                      ***/
/*** Copyright (C) Marcel de Kogel 1996,1997                              ***/
/***     You are not allowed to distribute this software commercially     ***/
/***     Please, notify me, if you make any changes to this file          ***/
/*** History:                                                             ***/
/*** 04/28/98 jb  Added DD/FD CB xx undoc'd opcodes (e.g. RLC B=(IX+n))   ***/
/***                                                                      ***/
/****************************************************************************/

// RpCׂ̈̕ύX狖ĉ܂

#include	<string.h>
#include	<stdlib.h>
#include	"Z80.h"
#include	"z80_2nd.h"

Z80OPCODE	Interrupt(void);
Z80OPCODE	ei(void);


		Z80_Regs	R;
		byte		Z80_Running = 1;
		word		Z80_ICount = 0;


#ifdef DEBUG
	static byte DEBUG_FLG=0;
	static byte TRACE_FLG=0;
	static byte Z80_Trace=0;
	static int  Z80_Trap=-1;
	static unsigned pc_trace[256];
	static unsigned pc_count=0;
	void Z80_Debug(Z80_Regs *R);
#endif

byte	PTable[512];
byte	ZSTable[512];
byte	ZSPTable[512];
#include "Z80DAA.h"

typedef void (*near opcode_fn)(void);

#define M_C     (R.AF.B.l&C_FLAG)
#define M_NC    (!M_C)
#define M_Z     (R.AF.B.l&Z_FLAG)
#define M_NZ    (!M_Z)
#define M_M     (R.AF.B.l&S_FLAG)
#define M_P     (!M_M)
#define M_PE    (R.AF.B.l&V_FLAG)
#define M_PO    (!M_PE)

/* Get next opcode argument and increment program counter */

static byte near M_RDMEM_OPCODE(void) {

	return(M_RDOP_ARG(R.PC.D++));
}

word M_RDMEM_OPCODE_WORD(void) {

	word	i;

	i = (word)M_RDOP_ARG(R.PC.D++);
	i += (word)M_RDOP_ARG(R.PC.D++) << 8;
	return(i);
}

static word M_RDMEM_WORD(word A) {

	word	i;

	i = (word)M_RDMEM(A++);
	i += (word)M_RDMEM(A) << 8;
	return(i);
}

static void M_WRMEM_WORD(word A, word V) {

	M_WRMEM(A++, (byte)(V & 0x00ff));
	M_WRMEM(A, (byte)(V >> 8));
}

#define M_XIX       (R.IX.D + cbw(M_RDMEM_OPCODE()))
#define M_XIY       (R.IY.D + cbw(M_RDMEM_OPCODE()))
#define M_RD_XHL    M_RDMEM(R.HL.D)
#define	M_RD_XIX	M_RDMEM(M_XIX)
#define	M_RD_XIY	M_RDMEM(M_XIY)

void M_WR_XIX(byte a) { M_WRMEM(M_XIX, a); }
void M_WR_XIY(byte a) { M_WRMEM(M_XIY, a); }

#include "Z80Codes.h"

Z80OPCODE adc_a_xhl(void) { byte i=M_RD_XHL; M_ADC(i); }
Z80OPCODE adc_a_xix(void) { byte i=M_RD_XIX; M_ADC(i); }
Z80OPCODE adc_a_xiy(void) { byte i=M_RD_XIY; M_ADC(i); }
Z80OPCODE adc_a_a(void) { M_ADC(R.AF.B.h); }
Z80OPCODE adc_a_b(void) { M_ADC(R.BC.B.h); }
Z80OPCODE adc_a_c(void) { M_ADC(R.BC.B.l); }
Z80OPCODE adc_a_d(void) { M_ADC(R.DE.B.h); }
Z80OPCODE adc_a_e(void) { M_ADC(R.DE.B.l); }
Z80OPCODE adc_a_h(void) { M_ADC(R.HL.B.h); }
Z80OPCODE adc_a_l(void) { M_ADC(R.HL.B.l); }
Z80OPCODE adc_a_ixl(void) { M_ADC(R.IX.B.l); }
Z80OPCODE adc_a_ixh(void) { M_ADC(R.IX.B.h); }
Z80OPCODE adc_a_iyl(void) { M_ADC(R.IY.B.l); }
Z80OPCODE adc_a_iyh(void) { M_ADC(R.IY.B.h); }
Z80OPCODE adc_a_byte(void) { byte i=M_RDMEM_OPCODE(); M_ADC(i); }

Z80OPCODE adc_hl_bc(void) { M_ADCW(BC); }
Z80OPCODE adc_hl_de(void) { M_ADCW(DE); }
Z80OPCODE adc_hl_hl(void) { M_ADCW(HL); }
Z80OPCODE adc_hl_sp(void) { M_ADCW(SP); }

Z80OPCODE add_a_xhl(void) { byte i=M_RD_XHL; M_ADD(i); }
Z80OPCODE add_a_xix(void) { byte i=M_RD_XIX; M_ADD(i); }
Z80OPCODE add_a_xiy(void) { byte i=M_RD_XIY; M_ADD(i); }
Z80OPCODE add_a_a(void) { M_ADD(R.AF.B.h); }
Z80OPCODE add_a_b(void) { M_ADD(R.BC.B.h); }
Z80OPCODE add_a_c(void) { M_ADD(R.BC.B.l); }
Z80OPCODE add_a_d(void) { M_ADD(R.DE.B.h); }
Z80OPCODE add_a_e(void) { M_ADD(R.DE.B.l); }
Z80OPCODE add_a_h(void) { M_ADD(R.HL.B.h); }
Z80OPCODE add_a_l(void) { M_ADD(R.HL.B.l); }
Z80OPCODE add_a_ixl(void) { M_ADD(R.IX.B.l); }
Z80OPCODE add_a_ixh(void) { M_ADD(R.IX.B.h); }
Z80OPCODE add_a_iyl(void) { M_ADD(R.IY.B.l); }
Z80OPCODE add_a_iyh(void) { M_ADD(R.IY.B.h); }
Z80OPCODE add_a_byte(void) { byte i = M_RDMEM_OPCODE(); M_ADD(i); }

Z80OPCODE add_hl_bc(void) { M_ADDW(HL,BC); }
Z80OPCODE add_hl_de(void) { M_ADDW(HL,DE); }
Z80OPCODE add_hl_hl(void) { M_ADDW(HL,HL); }
Z80OPCODE add_hl_sp(void) { M_ADDW(HL,SP); }
Z80OPCODE add_ix_bc(void) { M_ADDW(IX,BC); }
Z80OPCODE add_ix_de(void) { M_ADDW(IX,DE); }
Z80OPCODE add_ix_ix(void) { M_ADDW(IX,IX); }
Z80OPCODE add_ix_sp(void) { M_ADDW(IX,SP); }
Z80OPCODE add_iy_bc(void) { M_ADDW(IY,BC); }
Z80OPCODE add_iy_de(void) { M_ADDW(IY,DE); }
Z80OPCODE add_iy_iy(void) { M_ADDW(IY,IY); }
Z80OPCODE add_iy_sp(void) { M_ADDW(IY,SP); }

Z80OPCODE and_xhl(void) { byte i = M_RD_XHL; M_AND(i); }
Z80OPCODE and_xix(void) { byte i = M_RD_XIX; M_AND(i); }
Z80OPCODE and_xiy(void) { byte i = M_RD_XIY; M_AND(i); }
Z80OPCODE and_a(void) { R.AF.B.l = ZSPTable[R.AF.B.h] | (byte)H_FLAG; }
Z80OPCODE and_b(void) { M_AND(R.BC.B.h); }
Z80OPCODE and_c(void) { M_AND(R.BC.B.l); }
Z80OPCODE and_d(void) { M_AND(R.DE.B.h); }
Z80OPCODE and_e(void) { M_AND(R.DE.B.l); }
Z80OPCODE and_h(void) { M_AND(R.HL.B.h); }
Z80OPCODE and_l(void) { M_AND(R.HL.B.l); }
Z80OPCODE and_ixh(void) { M_AND(R.IX.B.h); }
Z80OPCODE and_ixl(void) { M_AND(R.IX.B.l); }
Z80OPCODE and_iyh(void) { M_AND(R.IY.B.h); }
Z80OPCODE and_iyl(void) { M_AND(R.IY.B.l); }
Z80OPCODE and_byte(void) { byte i = M_RDMEM_OPCODE(); M_AND(i); }


Z80OPCODE call_c(void) { if (M_C) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_m(void) { if (M_M) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_nc(void) { if (M_NC) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_nz(void) { if (M_NZ) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_p(void) { if (M_P) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_pe(void) { if (M_PE) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_po(void) { if (M_PO) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call_z(void) { if (M_Z) { M_CALL(); } else { M_SKIP_CALL; } }
Z80OPCODE call(void) { M_CALL(); }

#if 1											// '98/02/12
Z80OPCODE ccf(void) {
	R.AF.B.l = (byte)(((R.AF.B.l&0xED)|((R.AF.B.l&1)<<4))^1);
}
#else
Z80OPCODE ccf(void) { R.AF.B.l = (R.AF.B.l&0xED)|((R.AF.B.l&1)<<4)^1; }
#endif

Z80OPCODE cp_xhl(void) { byte i=M_RD_XHL; M_CP(i); }
Z80OPCODE cp_xix(void) { byte i=M_RD_XIX; M_CP(i); }
Z80OPCODE cp_xiy(void) { byte i=M_RD_XIY; M_CP(i); }
Z80OPCODE cp_a(void) { M_CP(R.AF.B.h); }
Z80OPCODE cp_b(void) { M_CP(R.BC.B.h); }
Z80OPCODE cp_c(void) { M_CP(R.BC.B.l); }
Z80OPCODE cp_d(void) { M_CP(R.DE.B.h); }
Z80OPCODE cp_e(void) { M_CP(R.DE.B.l); }
Z80OPCODE cp_h(void) { M_CP(R.HL.B.h); }
Z80OPCODE cp_l(void) { M_CP(R.HL.B.l); }
Z80OPCODE cp_ixh(void) { M_CP(R.IX.B.h); }
Z80OPCODE cp_ixl(void) { M_CP(R.IX.B.l); }
Z80OPCODE cp_iyh(void) { M_CP(R.IY.B.h); }
Z80OPCODE cp_iyl(void) { M_CP(R.IY.B.l); }
Z80OPCODE cp_byte(void) { byte i=M_RDMEM_OPCODE(); M_CP(i); }

Z80OPCODE cpd(void) {

	byte	i, j;

	i = M_RDMEM(R.HL.D);									// !!
	j = R.AF.B.h - i;
	--R.HL.W.l;
	--R.BC.W.l;
	R.AF.B.l = (byte)((R.AF.B.l&C_FLAG)|ZSTable[j]|
				((R.AF.B.h^i^j)&H_FLAG)|(R.BC.D? V_FLAG:0)|N_FLAG);
}

Z80OPCODE cpdr(void) {

	cpd();
	if (R.BC.D && !(R.AF.B.l&Z_FLAG)) { Z80_ICount+=5; R.PC.W.l-=2; }
}

Z80OPCODE cpi(void) {

	byte i,j;

	i = M_RDMEM(R.HL.D);									// !!
	j = R.AF.B.h - i;
	++R.HL.W.l;
	--R.BC.W.l;
	R.AF.B.l = (byte)((R.AF.B.l&C_FLAG)|ZSTable[j]|
				((R.AF.B.h^i^j)&H_FLAG)|(R.BC.D? V_FLAG:0)|N_FLAG);
}

Z80OPCODE cpir(void) {

	cpi();
	if (R.BC.D && !(R.AF.B.l&Z_FLAG)) { Z80_ICount+=5; R.PC.W.l-=2; }
}

Z80OPCODE cpl(void) { R.AF.B.h^=0xFF; R.AF.B.l|=(H_FLAG|N_FLAG); }

Z80OPCODE daa(void) {

	int i;											// int

	i = (int)R.AF.B.h;
	if (R.AF.B.l&C_FLAG) i|=256;
	if (R.AF.B.l&H_FLAG) i|=512;
	if (R.AF.B.l&N_FLAG) i|=1024;
	R.AF.W.l = DAATable[i];
}

Z80OPCODE dec_xhl(void) {

	byte	i;

	i = M_RDMEM(R.HL.D);
	M_DEC(i);
	M_WRMEM(R.HL.D,i);
}

Z80OPCODE dec_xix(void) {

	byte	i;
	int		j;

	j=M_XIX;
	i=M_RDMEM(j);
	M_DEC(i);
	M_WRMEM(j,i);
}
Z80OPCODE dec_xiy(void) {

	byte	i;
	int		j;

	j = M_XIY;
	i = M_RDMEM(j);
	M_DEC(i);
	M_WRMEM(j, i);
}

Z80OPCODE dec_a(void) { M_DEC(R.AF.B.h); }
Z80OPCODE dec_b(void) { M_DEC(R.BC.B.h); }
Z80OPCODE dec_c(void) { M_DEC(R.BC.B.l); }
Z80OPCODE dec_d(void) { M_DEC(R.DE.B.h); }
Z80OPCODE dec_e(void) { M_DEC(R.DE.B.l); }
Z80OPCODE dec_h(void) { M_DEC(R.HL.B.h); }
Z80OPCODE dec_l(void) { M_DEC(R.HL.B.l); }
Z80OPCODE dec_ixh(void) { M_DEC(R.IX.B.h); }
Z80OPCODE dec_ixl(void) { M_DEC(R.IX.B.l); }
Z80OPCODE dec_iyh(void) { M_DEC(R.IY.B.h); }
Z80OPCODE dec_iyl(void) { M_DEC(R.IY.B.l); }

Z80OPCODE dec_bc(void) { --R.BC.W.l; }
Z80OPCODE dec_de(void) { --R.DE.W.l; }
Z80OPCODE dec_hl(void) { --R.HL.W.l; }
Z80OPCODE dec_ix(void) { --R.IX.W.l; }
Z80OPCODE dec_iy(void) { --R.IY.W.l; }
Z80OPCODE dec_sp(void) { --R.SP.W.l; }

Z80OPCODE di(void) { R.IFF1=R.IFF2=0; }

Z80OPCODE djnz(void) { if (--R.BC.B.h) { M_JR(); } else { M_SKIP_JR; } }

Z80OPCODE ex_xsp_hl(void) {

	word	i;

	i = M_RDMEM_WORD(R.SP.D);
	M_WRMEM_WORD(R.SP.D, R.HL.D);
	R.HL.D = i;
}

Z80OPCODE ex_xsp_ix(void) {

	word	i;

	i = M_RDMEM_WORD(R.SP.D);
	M_WRMEM_WORD(R.SP.D, R.IX.D);
	R.IX.D = i;
}

Z80OPCODE ex_xsp_iy(void) {

	word	i;

	i = M_RDMEM_WORD(R.SP.D);
	M_WRMEM_WORD(R.SP.D, R.IY.D);
	R.IY.D = i;
}

Z80OPCODE ex_af_af(void) {

	word	i;

	i = R.AF.D;
	R.AF.D=R.AF2.D;
	R.AF2.D = i;
}

Z80OPCODE ex_de_hl(void) {

	word	i;

	i = R.DE.D;
	R.DE.D = R.HL.D;
	R.HL.D = i;
}

Z80OPCODE exx(void) {

	word	i;

	i = R.BC.D;
	R.BC.D = R.BC2.D;
	R.BC2.D = i;
	i = R.DE.D;
	R.DE.D = R.DE2.D;
	R.DE2.D = i;
	i = R.HL.D;
	R.HL.D = R.HL2.D;
	R.HL2.D = i;
}

Z80OPCODE halt(void) {

	R.PC.W.l--;
	R.HALT = 1;
}

Z80OPCODE im_0(void) { R.IM=0; }
Z80OPCODE im_1(void) { R.IM=1; }
Z80OPCODE im_2(void) { R.IM=2; }

Z80OPCODE in_a_c(void) { M_IN(R.AF.B.h); }
Z80OPCODE in_b_c(void) { M_IN(R.BC.B.h); }
Z80OPCODE in_c_c(void) { M_IN(R.BC.B.l); }
Z80OPCODE in_d_c(void) { M_IN(R.DE.B.h); }
Z80OPCODE in_e_c(void) { M_IN(R.DE.B.l); }
Z80OPCODE in_h_c(void) { M_IN(R.HL.B.h); }
Z80OPCODE in_l_c(void) { M_IN(R.HL.B.l); }
Z80OPCODE in_0_c(void) { byte i; M_IN(i); }

Z80OPCODE in_a_byte(void) {

	byte i = M_RDMEM_OPCODE();								// !!
	R.AF.B.h = DoIn(i, R.AF.B.h);
}

Z80OPCODE inc_xhl(void) {

	byte i;

	i=M_RDMEM(R.HL.D);
	M_INC(i);
	M_WRMEM(R.HL.D,i);
}

Z80OPCODE inc_xix(void) {

	byte	i;
	int		j;

	j = M_XIX;
	i = M_RDMEM(j);
	M_INC(i);
	M_WRMEM(j,i);
}
Z80OPCODE inc_xiy(void) {

	word	adr;
	byte	mem;

	adr = M_XIY;
	mem = M_RDMEM(adr);
	M_INC(mem);
	M_WRMEM(adr, mem);
}

Z80OPCODE inc_a(void) { M_INC(R.AF.B.h); }
Z80OPCODE inc_b(void) { M_INC(R.BC.B.h); }
Z80OPCODE inc_c(void) { M_INC(R.BC.B.l); }
Z80OPCODE inc_d(void) { M_INC(R.DE.B.h); }
Z80OPCODE inc_e(void) { M_INC(R.DE.B.l); }
Z80OPCODE inc_h(void) { M_INC(R.HL.B.h); }
Z80OPCODE inc_l(void) { M_INC(R.HL.B.l); }
Z80OPCODE inc_ixh(void) { M_INC(R.IX.B.h); }
Z80OPCODE inc_ixl(void) { M_INC(R.IX.B.l); }
Z80OPCODE inc_iyh(void) { M_INC(R.IY.B.h); }
Z80OPCODE inc_iyl(void) { M_INC(R.IY.B.l); }

Z80OPCODE inc_bc(void) { ++R.BC.W.l; }
Z80OPCODE inc_de(void) { ++R.DE.W.l; }
Z80OPCODE inc_hl(void) { ++R.HL.W.l; }
Z80OPCODE inc_ix(void) { ++R.IX.W.l; }
Z80OPCODE inc_iy(void) { ++R.IY.W.l; }
Z80OPCODE inc_sp(void) { ++R.SP.W.l; }

Z80OPCODE ind(void) {

#if 1
	M_WRMEM(R.HL.D, DoIn_N);
	--R.BC.B.h;
#else
	--R.BC.B.h;
	M_WRMEM(R.HL.D, DoIn_N);
#endif
	--R.HL.W.l;
	R.AF.B.l=(byte)((R.BC.B.h)? N_FLAG:(N_FLAG|Z_FLAG));
}

Z80OPCODE indr(void) {
	ind();
	if(R.BC.B.h){ Z80_ICount+=5; R.PC.W.l-=2; }
}

Z80OPCODE ini(void) {

#if 1											// !!! 990621
	M_WRMEM(R.HL.D, DoIn_N);
	--R.BC.B.h;
#else
	--R.BC.B.h;
	M_WRMEM(R.HL.D, DoIn_N);
#endif
	++R.HL.W.l;
	R.AF.B.l = (byte)((R.BC.B.h)? N_FLAG:(N_FLAG|Z_FLAG));
}

Z80OPCODE inir(void) {

	ini();
	if(R.BC.B.h) { Z80_ICount+=5; R.PC.W.l-=2; }
}
Z80OPCODE jp(void){ M_JP(); }
Z80OPCODE jp_hl(void) { R.PC.D=R.HL.D; }	/* TS 971002 */
Z80OPCODE jp_ix(void) { R.PC.D=R.IX.D; }	/* TS 971002 */
Z80OPCODE jp_iy(void) { R.PC.D=R.IY.D; }	/* TS 971002 */
Z80OPCODE jp_c(void) { if (M_C) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_m(void) { if (M_M) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_nc(void) { if (M_NC) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_nz(void) { if (M_NZ) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_p(void) { if (M_P) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_pe(void) { if (M_PE) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_po(void) { if (M_PO) { M_JP(); } else { M_SKIP_JP; } }
Z80OPCODE jp_z(void) { if (M_Z) { M_JP(); } else { M_SKIP_JP; } }

Z80OPCODE jr(void) { M_JR(); }
Z80OPCODE jr_c(void) { if (M_C) { M_JR(); } else { M_SKIP_JR; } }
Z80OPCODE jr_nc(void) { if (M_NC) { M_JR(); } else { M_SKIP_JR; } }
Z80OPCODE jr_nz(void) { if (M_NZ) { M_JR(); } else { M_SKIP_JR; } }
Z80OPCODE jr_z(void) { if (M_Z) { M_JR(); } else { M_SKIP_JR; } }

Z80OPCODE ld_xbc_a(void) { M_WRMEM(R.BC.D,R.AF.B.h); }
Z80OPCODE ld_xde_a(void) { M_WRMEM(R.DE.D,R.AF.B.h); }
Z80OPCODE ld_xhl_a(void) { M_WRMEM(R.HL.D,R.AF.B.h); }
Z80OPCODE ld_xhl_b(void) { M_WRMEM(R.HL.D,R.BC.B.h); }
Z80OPCODE ld_xhl_c(void) { M_WRMEM(R.HL.D,R.BC.B.l); }
Z80OPCODE ld_xhl_d(void) { M_WRMEM(R.HL.D,R.DE.B.h); }
Z80OPCODE ld_xhl_e(void) { M_WRMEM(R.HL.D,R.DE.B.l); }
Z80OPCODE ld_xhl_h(void) { M_WRMEM(R.HL.D,R.HL.B.h); }
Z80OPCODE ld_xhl_l(void) { M_WRMEM(R.HL.D,R.HL.B.l); }
Z80OPCODE ld_xhl_byte(void) { byte i=M_RDMEM_OPCODE(); M_WRMEM(R.HL.D,i); }
Z80OPCODE ld_xix_a(void) { M_WR_XIX(R.AF.B.h); }
Z80OPCODE ld_xix_b(void) { M_WR_XIX(R.BC.B.h); }
Z80OPCODE ld_xix_c(void) { M_WR_XIX(R.BC.B.l); }
Z80OPCODE ld_xix_d(void) { M_WR_XIX(R.DE.B.h); }
Z80OPCODE ld_xix_e(void) { M_WR_XIX(R.DE.B.l); }
Z80OPCODE ld_xix_h(void) { M_WR_XIX(R.HL.B.h); }
Z80OPCODE ld_xix_l(void) { M_WR_XIX(R.HL.B.l); }
Z80OPCODE ld_xix_byte(void) {

	int		i;
	byte	j;

	i = M_XIX;
	j = M_RDMEM_OPCODE();
	M_WRMEM(i, j);
}

Z80OPCODE ld_xiy_a(void) { M_WR_XIY(R.AF.B.h); }
Z80OPCODE ld_xiy_b(void) { M_WR_XIY(R.BC.B.h); }
Z80OPCODE ld_xiy_c(void) { M_WR_XIY(R.BC.B.l); }
Z80OPCODE ld_xiy_d(void) { M_WR_XIY(R.DE.B.h); }
Z80OPCODE ld_xiy_e(void) { M_WR_XIY(R.DE.B.l); }
Z80OPCODE ld_xiy_h(void) { M_WR_XIY(R.HL.B.h); }
Z80OPCODE ld_xiy_l(void) { M_WR_XIY(R.HL.B.l); }
Z80OPCODE ld_xiy_byte(void) {

	int		i;
	byte	j;

	i=M_XIY;
	j=M_RDMEM_OPCODE();
	M_WRMEM(i,j);
}

Z80OPCODE ld_xbyte_a(void) { int i=M_RDMEM_OPCODE_WORD(); M_WRMEM(i,R.AF.B.h); }
Z80OPCODE ld_xword_bc(void) { M_WRMEM_WORD(M_RDMEM_OPCODE_WORD(),R.BC.D); }
Z80OPCODE ld_xword_de(void) { M_WRMEM_WORD(M_RDMEM_OPCODE_WORD(),R.DE.D); }
Z80OPCODE ld_xword_hl(void) { M_WRMEM_WORD(M_RDMEM_OPCODE_WORD(),R.HL.D); }
Z80OPCODE ld_xword_ix(void) { M_WRMEM_WORD(M_RDMEM_OPCODE_WORD(),R.IX.D); }
Z80OPCODE ld_xword_iy(void) { M_WRMEM_WORD(M_RDMEM_OPCODE_WORD(),R.IY.D); }
Z80OPCODE ld_xword_sp(void) { M_WRMEM_WORD(M_RDMEM_OPCODE_WORD(),R.SP.D); }
Z80OPCODE ld_a_xbc(void) { R.AF.B.h=M_RDMEM(R.BC.D); }
Z80OPCODE ld_a_xde(void) { R.AF.B.h=M_RDMEM(R.DE.D); }
Z80OPCODE ld_a_xhl(void) { R.AF.B.h=M_RD_XHL; }
Z80OPCODE ld_a_xix(void) { R.AF.B.h=M_RD_XIX; }
Z80OPCODE ld_a_xiy(void) { R.AF.B.h=M_RD_XIY; }
Z80OPCODE ld_a_xbyte(void)
{ int i=M_RDMEM_OPCODE_WORD(); R.AF.B.h=M_RDMEM(i); }

Z80OPCODE ld_a_byte(void) { R.AF.B.h=M_RDMEM_OPCODE(); }
Z80OPCODE ld_b_byte(void) { R.BC.B.h=M_RDMEM_OPCODE(); }
Z80OPCODE ld_c_byte(void) { R.BC.B.l=M_RDMEM_OPCODE(); }
Z80OPCODE ld_d_byte(void) { R.DE.B.h=M_RDMEM_OPCODE(); }
Z80OPCODE ld_e_byte(void) { R.DE.B.l=M_RDMEM_OPCODE(); }
Z80OPCODE ld_h_byte(void) { R.HL.B.h=M_RDMEM_OPCODE(); }
Z80OPCODE ld_l_byte(void) { R.HL.B.l=M_RDMEM_OPCODE(); }
Z80OPCODE ld_ixh_byte(void) { R.IX.B.h=M_RDMEM_OPCODE(); }
Z80OPCODE ld_ixl_byte(void) { R.IX.B.l=M_RDMEM_OPCODE(); }
Z80OPCODE ld_iyh_byte(void) { R.IY.B.h=M_RDMEM_OPCODE(); }
Z80OPCODE ld_iyl_byte(void) { R.IY.B.l=M_RDMEM_OPCODE(); }

Z80OPCODE ld_b_xhl(void) { R.BC.B.h=M_RD_XHL; }
Z80OPCODE ld_c_xhl(void) { R.BC.B.l=M_RD_XHL; }
Z80OPCODE ld_d_xhl(void) { R.DE.B.h=M_RD_XHL; }
Z80OPCODE ld_e_xhl(void) { R.DE.B.l=M_RD_XHL; }
Z80OPCODE ld_h_xhl(void) { R.HL.B.h=M_RD_XHL; }
Z80OPCODE ld_l_xhl(void) { R.HL.B.l=M_RD_XHL; }
Z80OPCODE ld_b_xix(void) { R.BC.B.h=M_RD_XIX; }
Z80OPCODE ld_c_xix(void) { R.BC.B.l=M_RD_XIX; }
Z80OPCODE ld_d_xix(void) { R.DE.B.h=M_RD_XIX; }
Z80OPCODE ld_e_xix(void) { R.DE.B.l=M_RD_XIX; }
Z80OPCODE ld_h_xix(void) { R.HL.B.h=M_RD_XIX; }
Z80OPCODE ld_l_xix(void) { R.HL.B.l=M_RD_XIX; }
Z80OPCODE ld_b_xiy(void) { R.BC.B.h=M_RD_XIY; }
Z80OPCODE ld_c_xiy(void) { R.BC.B.l=M_RD_XIY; }
Z80OPCODE ld_d_xiy(void) { R.DE.B.h=M_RD_XIY; }
Z80OPCODE ld_e_xiy(void) { R.DE.B.l=M_RD_XIY; }
Z80OPCODE ld_h_xiy(void) { R.HL.B.h=M_RD_XIY; }
Z80OPCODE ld_l_xiy(void) { R.HL.B.l=M_RD_XIY; }
Z80OPCODE ld_a_a(void) { }
Z80OPCODE ld_a_b(void) { R.AF.B.h=R.BC.B.h; }
Z80OPCODE ld_a_c(void) { R.AF.B.h=R.BC.B.l; }
Z80OPCODE ld_a_d(void) { R.AF.B.h=R.DE.B.h; }
Z80OPCODE ld_a_e(void) { R.AF.B.h=R.DE.B.l; }
Z80OPCODE ld_a_h(void) { R.AF.B.h=R.HL.B.h; }
Z80OPCODE ld_a_l(void) { R.AF.B.h=R.HL.B.l; }
Z80OPCODE ld_a_ixh(void) { R.AF.B.h=R.IX.B.h; }
Z80OPCODE ld_a_ixl(void) { R.AF.B.h=R.IX.B.l; }
Z80OPCODE ld_a_iyh(void) { R.AF.B.h=R.IY.B.h; }
Z80OPCODE ld_a_iyl(void) { R.AF.B.h=R.IY.B.l; }
Z80OPCODE ld_b_b(void) { }
Z80OPCODE ld_b_a(void) { R.BC.B.h=R.AF.B.h; }
Z80OPCODE ld_b_c(void) { R.BC.B.h=R.BC.B.l; }
Z80OPCODE ld_b_d(void) { R.BC.B.h=R.DE.B.h; }
Z80OPCODE ld_b_e(void) { R.BC.B.h=R.DE.B.l; }
Z80OPCODE ld_b_h(void) { R.BC.B.h=R.HL.B.h; }
Z80OPCODE ld_b_l(void) { R.BC.B.h=R.HL.B.l; }
Z80OPCODE ld_b_ixh(void) { R.BC.B.h=R.IX.B.h; }
Z80OPCODE ld_b_ixl(void) { R.BC.B.h=R.IX.B.l; }
Z80OPCODE ld_b_iyh(void) { R.BC.B.h=R.IY.B.h; }
Z80OPCODE ld_b_iyl(void) { R.BC.B.h=R.IY.B.l; }
Z80OPCODE ld_c_c(void) { }
Z80OPCODE ld_c_a(void) { R.BC.B.l=R.AF.B.h; }
Z80OPCODE ld_c_b(void) { R.BC.B.l=R.BC.B.h; }
Z80OPCODE ld_c_d(void) { R.BC.B.l=R.DE.B.h; }
Z80OPCODE ld_c_e(void) { R.BC.B.l=R.DE.B.l; }
Z80OPCODE ld_c_h(void) { R.BC.B.l=R.HL.B.h; }
Z80OPCODE ld_c_l(void) { R.BC.B.l=R.HL.B.l; }
Z80OPCODE ld_c_ixh(void) { R.BC.B.l=R.IX.B.h; }
Z80OPCODE ld_c_ixl(void) { R.BC.B.l=R.IX.B.l; }
Z80OPCODE ld_c_iyh(void) { R.BC.B.l=R.IY.B.h; }
Z80OPCODE ld_c_iyl(void) { R.BC.B.l=R.IY.B.l; }
Z80OPCODE ld_d_d(void) { }
Z80OPCODE ld_d_a(void) { R.DE.B.h=R.AF.B.h; }
Z80OPCODE ld_d_c(void) { R.DE.B.h=R.BC.B.l; }
Z80OPCODE ld_d_b(void) { R.DE.B.h=R.BC.B.h; }
Z80OPCODE ld_d_e(void) { R.DE.B.h=R.DE.B.l; }
Z80OPCODE ld_d_h(void) { R.DE.B.h=R.HL.B.h; }
Z80OPCODE ld_d_l(void) { R.DE.B.h=R.HL.B.l; }
Z80OPCODE ld_d_ixh(void) { R.DE.B.h=R.IX.B.h; }
Z80OPCODE ld_d_ixl(void) { R.DE.B.h=R.IX.B.l; }
Z80OPCODE ld_d_iyh(void) { R.DE.B.h=R.IY.B.h; }
Z80OPCODE ld_d_iyl(void) { R.DE.B.h=R.IY.B.l; }
Z80OPCODE ld_e_e(void) { }
Z80OPCODE ld_e_a(void) { R.DE.B.l=R.AF.B.h; }
Z80OPCODE ld_e_c(void) { R.DE.B.l=R.BC.B.l; }
Z80OPCODE ld_e_b(void) { R.DE.B.l=R.BC.B.h; }
Z80OPCODE ld_e_d(void) { R.DE.B.l=R.DE.B.h; }
Z80OPCODE ld_e_h(void) { R.DE.B.l=R.HL.B.h; }
Z80OPCODE ld_e_l(void) { R.DE.B.l=R.HL.B.l; }
Z80OPCODE ld_e_ixh(void) { R.DE.B.l=R.IX.B.h; }
Z80OPCODE ld_e_ixl(void) { R.DE.B.l=R.IX.B.l; }
Z80OPCODE ld_e_iyh(void) { R.DE.B.l=R.IY.B.h; }
Z80OPCODE ld_e_iyl(void) { R.DE.B.l=R.IY.B.l; }
Z80OPCODE ld_h_h(void) { }
Z80OPCODE ld_h_a(void) { R.HL.B.h=R.AF.B.h; }
Z80OPCODE ld_h_c(void) { R.HL.B.h=R.BC.B.l; }
Z80OPCODE ld_h_b(void) { R.HL.B.h=R.BC.B.h; }
Z80OPCODE ld_h_e(void) { R.HL.B.h=R.DE.B.l; }
Z80OPCODE ld_h_d(void) { R.HL.B.h=R.DE.B.h; }
Z80OPCODE ld_h_l(void) { R.HL.B.h=R.HL.B.l; }
Z80OPCODE ld_l_l(void) { }
Z80OPCODE ld_l_a(void) { R.HL.B.l=R.AF.B.h; }
Z80OPCODE ld_l_c(void) { R.HL.B.l=R.BC.B.l; }
Z80OPCODE ld_l_b(void) { R.HL.B.l=R.BC.B.h; }
Z80OPCODE ld_l_e(void) { R.HL.B.l=R.DE.B.l; }
Z80OPCODE ld_l_d(void) { R.HL.B.l=R.DE.B.h; }
Z80OPCODE ld_l_h(void) { R.HL.B.l=R.HL.B.h; }
Z80OPCODE ld_ixh_a(void) { R.IX.B.h=R.AF.B.h; }
Z80OPCODE ld_ixh_b(void) { R.IX.B.h=R.BC.B.h; }
Z80OPCODE ld_ixh_c(void) { R.IX.B.h=R.BC.B.l; }
Z80OPCODE ld_ixh_d(void) { R.IX.B.h=R.DE.B.h; }
Z80OPCODE ld_ixh_e(void) { R.IX.B.h=R.DE.B.l; }
Z80OPCODE ld_ixh_ixh(void) { }
Z80OPCODE ld_ixh_ixl(void) { R.IX.B.h=R.IX.B.l; }
Z80OPCODE ld_ixl_a(void) { R.IX.B.l=R.AF.B.h; }
Z80OPCODE ld_ixl_b(void) { R.IX.B.l=R.BC.B.h; }
Z80OPCODE ld_ixl_c(void) { R.IX.B.l=R.BC.B.l; }
Z80OPCODE ld_ixl_d(void) { R.IX.B.l=R.DE.B.h; }
Z80OPCODE ld_ixl_e(void) { R.IX.B.l=R.DE.B.l; }
Z80OPCODE ld_ixl_ixh(void) { R.IX.B.l=R.IX.B.h; }
Z80OPCODE ld_ixl_ixl(void) { }
Z80OPCODE ld_iyh_a(void) { R.IY.B.h=R.AF.B.h; }
Z80OPCODE ld_iyh_b(void) { R.IY.B.h=R.BC.B.h; }
Z80OPCODE ld_iyh_c(void) { R.IY.B.h=R.BC.B.l; }
Z80OPCODE ld_iyh_d(void) { R.IY.B.h=R.DE.B.h; }
Z80OPCODE ld_iyh_e(void) { R.IY.B.h=R.DE.B.l; }
Z80OPCODE ld_iyh_iyh(void) { }
Z80OPCODE ld_iyh_iyl(void) { R.IY.B.h=R.IY.B.l; }
Z80OPCODE ld_iyl_a(void) { R.IY.B.l=R.AF.B.h; }
Z80OPCODE ld_iyl_b(void) { R.IY.B.l=R.BC.B.h; }
Z80OPCODE ld_iyl_c(void) { R.IY.B.l=R.BC.B.l; }
Z80OPCODE ld_iyl_d(void) { R.IY.B.l=R.DE.B.h; }
Z80OPCODE ld_iyl_e(void) { R.IY.B.l=R.DE.B.l; }
Z80OPCODE ld_iyl_iyh(void) { R.IY.B.l=R.IY.B.h; }
Z80OPCODE ld_iyl_iyl(void) { }
Z80OPCODE ld_bc_xword(void) { R.BC.D=M_RDMEM_WORD(M_RDMEM_OPCODE_WORD()); }
Z80OPCODE ld_bc_word(void) { R.BC.D=M_RDMEM_OPCODE_WORD(); }
Z80OPCODE ld_de_xword(void) { R.DE.D=M_RDMEM_WORD(M_RDMEM_OPCODE_WORD()); }
Z80OPCODE ld_de_word(void) { R.DE.D=M_RDMEM_OPCODE_WORD(); }
Z80OPCODE ld_hl_xword(void) { R.HL.D=M_RDMEM_WORD(M_RDMEM_OPCODE_WORD()); }
Z80OPCODE ld_hl_word(void) { R.HL.D=M_RDMEM_OPCODE_WORD(); }
Z80OPCODE ld_ix_xword(void) { R.IX.D=M_RDMEM_WORD(M_RDMEM_OPCODE_WORD()); }
Z80OPCODE ld_ix_word(void) { R.IX.D=M_RDMEM_OPCODE_WORD(); }
Z80OPCODE ld_iy_xword(void) { R.IY.D=M_RDMEM_WORD(M_RDMEM_OPCODE_WORD()); }
Z80OPCODE ld_iy_word(void) { R.IY.D=M_RDMEM_OPCODE_WORD(); }
Z80OPCODE ld_sp_xword(void) { R.SP.D=M_RDMEM_WORD(M_RDMEM_OPCODE_WORD()); }
Z80OPCODE ld_sp_word(void) { R.SP.D=M_RDMEM_OPCODE_WORD(); }
Z80OPCODE ld_sp_hl(void) { R.SP.D=R.HL.D; }
Z80OPCODE ld_sp_ix(void) { R.SP.D=R.IX.D; }
Z80OPCODE ld_sp_iy(void) { R.SP.D=R.IY.D; }
Z80OPCODE ld_a_i(void) {
	R.AF.B.h = (byte)R.I;
	R.AF.B.l = (byte)((R.AF.B.l&C_FLAG)|ZSTable[R.I]|(R.IFF2<<2));
}
Z80OPCODE ld_i_a(void) { R.I=R.AF.B.h; }
Z80OPCODE ld_a_r(void) {

	R.AF.B.h = (byte)((R.R&127)|(R.R2&128));
	R.AF.B.l = (byte)((R.AF.B.l&C_FLAG)|ZSTable[R.AF.B.h]|(R.IFF2<<2));
}

Z80OPCODE ld_r_a(void) { R.R = R.R2 = R.AF.B.h; }

Z80OPCODE ldd(void) {

	M_WRMEM(R.DE.D,M_RDMEM(R.HL.D));
	--R.DE.W.l;
	--R.HL.W.l;
	--R.BC.W.l;
	R.AF.B.l = (byte)((R.AF.B.l&0xE9)|(R.BC.D? V_FLAG:0));
}
Z80OPCODE lddr(void) {

	ldd();
	if (R.BC.D) { Z80_ICount+=5; R.PC.W.l-=2; }
}
Z80OPCODE ldi(void) {

	M_WRMEM(R.DE.D, M_RDMEM(R.HL.D));
	++R.DE.W.l;
	++R.HL.W.l;
	--R.BC.W.l;
	R.AF.B.l = (byte)((R.AF.B.l&0xE9)|(R.BC.D? V_FLAG:0));
}
Z80OPCODE ldir(void) {

	ldi();
	if (R.BC.D) { Z80_ICount+=5; R.PC.W.l-=2; }
}

Z80OPCODE neg(void)
{
 byte i;
 i=R.AF.B.h;
 R.AF.B.h=0;
 M_SUB(i);
}

Z80OPCODE nop(void) { }

Z80OPCODE or_xhl(void) { byte i=M_RD_XHL; M_OR(i); }
Z80OPCODE or_xix(void) { byte i=M_RD_XIX; M_OR(i); }
Z80OPCODE or_xiy(void) { byte i=M_RD_XIY; M_OR(i); }
Z80OPCODE or_a(void) { R.AF.B.l=ZSPTable[R.AF.B.h]; }
Z80OPCODE or_b(void) { M_OR(R.BC.B.h); }
Z80OPCODE or_c(void) { M_OR(R.BC.B.l); }
Z80OPCODE or_d(void) { M_OR(R.DE.B.h); }
Z80OPCODE or_e(void) { M_OR(R.DE.B.l); }
Z80OPCODE or_h(void) { M_OR(R.HL.B.h); }
Z80OPCODE or_l(void) { M_OR(R.HL.B.l); }
Z80OPCODE or_ixh(void) { M_OR(R.IX.B.h); }
Z80OPCODE or_ixl(void) { M_OR(R.IX.B.l); }
Z80OPCODE or_iyh(void) { M_OR(R.IY.B.h); }
Z80OPCODE or_iyl(void) { M_OR(R.IY.B.l); }
Z80OPCODE or_byte(void) { byte i=M_RDMEM_OPCODE(); M_OR(i); }

Z80OPCODE outd(void) {

 --R.BC.B.h;
 DoOut_N(M_RDMEM(R.HL.D));
 --R.HL.W.l;
 R.AF.B.l = (byte)(((R.BC.B.h)? N_FLAG:(Z_FLAG|N_FLAG)));
}

Z80OPCODE otdr(void) {

 outd();
 if (R.BC.B.h) { Z80_ICount+=5; R.PC.W.l-=2; }
}

Z80OPCODE outi(void) {

	--R.BC.B.h;
	DoOut_N(M_RDMEM(R.HL.D));
	++R.HL.W.l;
	R.AF.B.l = (byte)((R.BC.B.h)? N_FLAG:(Z_FLAG|N_FLAG));
}


Z80OPCODE otir(void) {

	outi();
	if (R.BC.B.h) { Z80_ICount+=5; R.PC.W.l-=2; }
}


Z80OPCODE out_c_a(void) { DoOut_N(R.AF.B.h); }
Z80OPCODE out_c_b(void) { DoOut_N(R.BC.B.h); }
Z80OPCODE out_c_c(void) { DoOut_N(R.BC.B.l); }
Z80OPCODE out_c_d(void) { DoOut_N(R.DE.B.h); }
Z80OPCODE out_c_e(void) { DoOut_N(R.DE.B.l); }
Z80OPCODE out_c_h(void) { DoOut_N(R.HL.B.h); }
Z80OPCODE out_c_l(void) { DoOut_N(R.HL.B.l); }
Z80OPCODE out_c_0(void) { DoOut_N(0); }
Z80OPCODE out_byte_a(void) {

	byte i=M_RDMEM_OPCODE();
	DoOut(i,R.AF.B.h,R.AF.B.h);
}

Z80OPCODE pop_af(void) { M_POP(&R.AF.D); }
Z80OPCODE pop_bc(void) { M_POP(&R.BC.D); }
Z80OPCODE pop_de(void) { M_POP(&R.DE.D); }
Z80OPCODE pop_hl(void) { M_POP(&R.HL.D); }
Z80OPCODE pop_ix(void) { M_POP(&R.IX.D); }
Z80OPCODE pop_iy(void) { M_POP(&R.IY.D); }

Z80OPCODE push_af(void) { M_PUSH(R.AF.D); }
Z80OPCODE push_bc(void) { M_PUSH(R.BC.D); }
Z80OPCODE push_de(void) { M_PUSH(R.DE.D); }
Z80OPCODE push_hl(void) { M_PUSH(R.HL.D); }
Z80OPCODE push_ix(void) { M_PUSH(R.IX.D); }
Z80OPCODE push_iy(void) { M_PUSH(R.IY.D); }


Z80OPCODE ret(void) { M_RET(); }
Z80OPCODE ret_c(void) { if (M_C) { M_RET(); } }
Z80OPCODE ret_m(void) { if (M_M) { M_RET(); } }
Z80OPCODE ret_nc(void) { if (M_NC) { M_RET(); } }
Z80OPCODE ret_nz(void) { if (M_NZ) { M_RET(); } }
Z80OPCODE ret_p(void) { if (M_P) { M_RET(); } }
Z80OPCODE ret_pe(void) { if (M_PE) { M_RET(); } }
Z80OPCODE ret_po(void) { if (M_PO) { M_RET(); } }
Z80OPCODE ret_z(void) { if (M_Z) { M_RET(); } }

Z80OPCODE reti(void)
{
	// Z80_Reti();
	M_RET();
}
Z80OPCODE retn(void)
{
	R.IFF1=R.IFF2;
	// Z80_Retn();
	M_RET();
}


Z80OPCODE rla(void){ M_RLA(); }


Z80OPCODE rlca(void)  { M_RLCA(); }

Z80OPCODE rld(void)
{
 byte i;
 i=M_RDMEM(R.HL.D);
 M_WRMEM(R.HL.D,(byte)(((word)i<<4)|(R.AF.B.h&0x0F)));
 R.AF.B.h = (byte)((R.AF.B.h&0xF0)|(i>>4));
 R.AF.B.l = (byte)(R.AF.B.l&C_FLAG)|ZSPTable[R.AF.B.h];
}


Z80OPCODE rra(void)  { M_RRA(); }


Z80OPCODE rrca(void)  { M_RRCA(); }

Z80OPCODE rrd(void)
{
 byte i;
 i=M_RDMEM(R.HL.D);
 M_WRMEM(R.HL.D,(byte)((i>>4)|((word)R.AF.B.h<<4)));
 R.AF.B.h = (byte)((R.AF.B.h&0xF0)|(i&0x0F));
 R.AF.B.l = (byte)(R.AF.B.l&C_FLAG)|ZSPTable[R.AF.B.h];
}

Z80OPCODE rst_00(void) { M_RST(0x00); }
Z80OPCODE rst_08(void) { M_RST(0x08); }
Z80OPCODE rst_10(void) { M_RST(0x10); }
Z80OPCODE rst_18(void) { M_RST(0x18); }
Z80OPCODE rst_20(void) { M_RST(0x20); }
Z80OPCODE rst_28(void) { M_RST(0x28); }
Z80OPCODE rst_30(void) { M_RST(0x30); }
Z80OPCODE rst_38(void) { M_RST(0x38); }

Z80OPCODE sbc_a_byte(void) { byte i=M_RDMEM_OPCODE(); M_SBC(i); }
Z80OPCODE sbc_a_xhl(void) { byte i=M_RD_XHL; M_SBC(i); }
Z80OPCODE sbc_a_xix(void) { byte i=M_RD_XIX; M_SBC(i); }
Z80OPCODE sbc_a_xiy(void) { byte i=M_RD_XIY; M_SBC(i); }
Z80OPCODE sbc_a_a(void) { M_SBC(R.AF.B.h); }
Z80OPCODE sbc_a_b(void) { M_SBC(R.BC.B.h); }
Z80OPCODE sbc_a_c(void) { M_SBC(R.BC.B.l); }
Z80OPCODE sbc_a_d(void) { M_SBC(R.DE.B.h); }
Z80OPCODE sbc_a_e(void) { M_SBC(R.DE.B.l); }
Z80OPCODE sbc_a_h(void) { M_SBC(R.HL.B.h); }
Z80OPCODE sbc_a_l(void) { M_SBC(R.HL.B.l); }
Z80OPCODE sbc_a_ixh(void) { M_SBC(R.IX.B.h); }
Z80OPCODE sbc_a_ixl(void) { M_SBC(R.IX.B.l); }
Z80OPCODE sbc_a_iyh(void) { M_SBC(R.IY.B.h); }
Z80OPCODE sbc_a_iyl(void) { M_SBC(R.IY.B.l); }

Z80OPCODE sbc_hl_bc(void) { M_SBCW(BC); }
Z80OPCODE sbc_hl_de(void) { M_SBCW(DE); }
Z80OPCODE sbc_hl_hl(void) { M_SBCW(HL); }
Z80OPCODE sbc_hl_sp(void) { M_SBCW(SP); }

Z80OPCODE scf(void) { R.AF.B.l=(byte)((R.AF.B.l&0xEC)|C_FLAG); }




Z80OPCODE sub_xhl(void) { byte i=M_RD_XHL; M_SUB(i); }
Z80OPCODE sub_xix(void) { byte i=M_RD_XIX; M_SUB(i); }
Z80OPCODE sub_xiy(void) { byte i=M_RD_XIY; M_SUB(i); }
Z80OPCODE sub_a(void) { R.AF.D=Z_FLAG|N_FLAG; }
Z80OPCODE sub_b(void) { M_SUB(R.BC.B.h); }
Z80OPCODE sub_c(void) { M_SUB(R.BC.B.l); }
Z80OPCODE sub_d(void) { M_SUB(R.DE.B.h); }
Z80OPCODE sub_e(void) { M_SUB(R.DE.B.l); }
Z80OPCODE sub_h(void) { M_SUB(R.HL.B.h); }
Z80OPCODE sub_l(void) { M_SUB(R.HL.B.l); }
Z80OPCODE sub_ixh(void) { M_SUB(R.IX.B.h); }
Z80OPCODE sub_ixl(void) { M_SUB(R.IX.B.l); }
Z80OPCODE sub_iyh(void) { M_SUB(R.IY.B.h); }
Z80OPCODE sub_iyl(void) { M_SUB(R.IY.B.l); }
Z80OPCODE sub_byte(void) { byte i=M_RDMEM_OPCODE(); M_SUB(i); }

Z80OPCODE xor_xhl(void) { byte i=M_RD_XHL; M_XOR(i); }
Z80OPCODE xor_xix(void) { byte i=M_RD_XIX; M_XOR(i); }
Z80OPCODE xor_xiy(void) { byte i=M_RD_XIY; M_XOR(i); }
Z80OPCODE xor_a(void) { R.AF.D=Z_FLAG|V_FLAG; }
Z80OPCODE xor_b(void) { M_XOR(R.BC.B.h); }
Z80OPCODE xor_c(void) { M_XOR(R.BC.B.l); }
Z80OPCODE xor_d(void) { M_XOR(R.DE.B.h); }
Z80OPCODE xor_e(void) { M_XOR(R.DE.B.l); }
Z80OPCODE xor_h(void) { M_XOR(R.HL.B.h); }
Z80OPCODE xor_l(void) { M_XOR(R.HL.B.l); }
Z80OPCODE xor_ixh(void) { M_XOR(R.IX.B.h); }
Z80OPCODE xor_ixl(void) { M_XOR(R.IX.B.l); }
Z80OPCODE xor_iyh(void) { M_XOR(R.IY.B.h); }
Z80OPCODE xor_iyl(void) { M_XOR(R.IY.B.l); }
Z80OPCODE xor_byte(void) { byte i=M_RDMEM_OPCODE(); M_XOR(i); }

Z80OPCODE no_op(void) { --R.PC.W.l; }

static byte cycles_main[256]={
  4,10,7,6,4,4,7,4,
  4,11,7,6,4,4,7,4,
  8,10,7,6,4,4,7,4,
  7,11,7,6,4,4,7,4,
  7,10,16,6,4,4,7,4,
  7,11,16,6,4,4,7,4,
  7,10,13,6,11,11,10,4,
  7,11,13,6,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  7,7,7,7,7,7,4,7,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  4,4,4,4,4,4,7,4,
  5,10,10,10,10,11,7,11,
  5,4,10,0,10,10,7,11,
  5,10,10,11,10,11,7,11,
  5,4,10,11,10,0,7,11,
  5,10,10,19,10,11,7,11,
  5,4,10,4,10,0,7,11,
  5,10,10,4,10,11,7,11,
  5,6,10,4,10,0,7,11
};

static byte cycles_cb[256]={
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,12,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8,
  8,8,8,8,8,8,15,8
};

static byte cycles_xx_cb[]={
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  20,20,20,20,20,20,20,20,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23,
  23,23,23,23,23,23,23,23
};

static byte cycles_xx[256]={
  0,0,0,0,0,0,0,0,
  0,15,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,15,0,0,0,0,0,0,
  0,14,20,10,9,9,9,0,
  0,15,20,10,9,9,9,0,
  0,0,0,0,23,23,19,0,
  0,15,0,0,0,0,0,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  9,9,9,9,9,9,19,9,	/* ASG 220997 */
  9,9,9,9,9,9,19,9,	/* ASG 220997 */
  19,19,19,19,19,19,19,19,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,9,9,19,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,14,0,23,0,15,0,0,
  0,8,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,10,0,0,0,0,0,0
};

static byte cycles_ed[256]={
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  12,12,15,20,8,8,8,9,
  12,12,15,20,8,8,8,9,
  12,12,15,20,8,8,8,9,
  12,12,15,20,8,8,8,9,
  12,12,15,20,8,8,8,18,
  12,12,15,20,8,8,8,18,
  12,12,15,20,8,8,8,0,
  12,12,15,20,8,8,8,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  16,16,16,16,0,0,0,0,
  16,16,16,16,0,0,0,0,
  16,16,16,16,0,0,0,0,  /* ASG 220997 */
  16,16,16,16,0,0,0,0,	/* ASG 220997 */
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0
};

Z80OPCODE no_op_xx(void) { R.PC.W.l++; }



Z80OPCODE dd_cb(void) {

	byte	opcode1;
	byte	opcode2;

	opcode1 = M_RDMEM_OPCODE();
	opcode2 = M_RDMEM_OPCODE();
	Z80_ICount += cycles_xx_cb[opcode2];
	proc_dd_cb(opcode1, opcode2);
}

Z80OPCODE fd_cb(void) {

	byte	opcode1;
	byte	opcode2;

	opcode1 = M_RDMEM_OPCODE();
	opcode2 = M_RDMEM_OPCODE();
	Z80_ICount += cycles_xx_cb[opcode2];
	proc_fd_cb(opcode1, opcode2);
}





static opcode_fn opcode_dd[256]=
{
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,add_ix_bc ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,add_ix_de ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,ld_ix_word,ld_xword_ix,inc_ix   ,inc_ixh    ,dec_ixh    ,ld_ixh_byte,no_op   ,
  no_op   ,add_ix_ix ,ld_ix_xword,dec_ix   ,inc_ixl    ,dec_ixl    ,ld_ixl_byte,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,inc_xix    ,dec_xix    ,ld_xix_byte,no_op   ,
  no_op   ,add_ix_sp ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_b_ixh   ,ld_b_ixl   ,ld_b_xix   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_c_ixh   ,ld_c_ixl   ,ld_c_xix   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_d_ixh   ,ld_d_ixl   ,ld_d_xix   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_e_ixh   ,ld_e_ixl   ,ld_e_xix   ,no_op   ,
  ld_ixh_b,ld_ixh_c  ,ld_ixh_d   ,ld_ixh_e ,ld_ixh_ixh ,ld_ixh_ixl ,ld_h_xix   ,ld_ixh_a,
  ld_ixl_b,ld_ixl_c  ,ld_ixl_d   ,ld_ixl_e ,ld_ixl_ixh ,ld_ixl_ixl ,ld_l_xix   ,ld_ixl_a,
  ld_xix_b,ld_xix_c  ,ld_xix_d   ,ld_xix_e ,ld_xix_h   ,ld_xix_l   ,no_op      ,ld_xix_a,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_a_ixh   ,ld_a_ixl   ,ld_a_xix   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,add_a_ixh  ,add_a_ixl  ,add_a_xix  ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,adc_a_ixh  ,adc_a_ixl  ,adc_a_xix  ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,sub_ixh    ,sub_ixl    ,sub_xix    ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,sbc_a_ixh  ,sbc_a_ixl  ,sbc_a_xix  ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,and_ixh    ,and_ixl    ,and_xix    ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,xor_ixh    ,xor_ixl    ,xor_xix    ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,or_ixh     ,or_ixl     ,or_xix     ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,cp_ixh     ,cp_ixl     ,cp_xix     ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,dd_cb    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,pop_ix    ,no_op      ,ex_xsp_ix,no_op      ,push_ix    ,no_op      ,no_op   ,
  no_op   ,jp_ix     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,ld_sp_ix  ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op
};

static opcode_fn opcode_ed[256]=
{
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 in_b_c,out_c_b,sbc_hl_bc,ld_xword_bc,neg,retn,im_0 ,ld_i_a,
 in_c_c,out_c_c,adc_hl_bc,ld_bc_xword,neg,reti,im_0 ,ld_r_a,
 in_d_c,out_c_d,sbc_hl_de,ld_xword_de,neg,retn,im_1 ,ld_a_i,
 in_e_c,out_c_e,adc_hl_de,ld_de_xword,neg,reti,im_2 ,ld_a_r,
 in_h_c,out_c_h,sbc_hl_hl,ld_xword_hl,neg,retn,im_0 ,rrd   ,
 in_l_c,out_c_l,adc_hl_hl,ld_hl_xword,neg,reti,im_0 ,rld   ,
 in_0_c,out_c_0,sbc_hl_sp,ld_xword_sp,neg,retn,im_1 ,nop   ,
 in_a_c,out_c_a,adc_hl_sp,ld_sp_xword,neg,reti,im_2 ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 ldi   ,cpi    ,ini      ,outi       ,nop,nop ,nop  ,nop   ,
 ldd   ,cpd    ,ind      ,outd       ,nop,nop ,nop  ,nop   ,
 ldir  ,cpir   ,inir     ,otir       ,nop,nop ,nop  ,nop   ,
 lddr  ,cpdr   ,indr     ,otdr       ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop   ,
 nop   ,nop    ,nop      ,nop        ,nop,nop ,nop  ,nop
//nop  ,nop    ,nop      ,nop        ,nop,nop ,patch,nop
};

static opcode_fn opcode_fd[256]=
{
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,add_iy_bc ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,add_iy_de ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,ld_iy_word,ld_xword_iy,inc_iy   ,inc_iyh    ,dec_iyh    ,ld_iyh_byte,no_op   ,
  no_op   ,add_iy_iy ,ld_iy_xword,dec_iy   ,inc_iyl    ,dec_iyl    ,ld_iyl_byte,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,inc_xiy    ,dec_xiy    ,ld_xiy_byte,no_op   ,
  no_op   ,add_iy_sp ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_b_iyh   ,ld_b_iyl   ,ld_b_xiy   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_c_iyh   ,ld_c_iyl   ,ld_c_xiy   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_d_iyh   ,ld_d_iyl   ,ld_d_xiy   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_e_iyh   ,ld_e_iyl   ,ld_e_xiy   ,no_op   ,
  ld_iyh_b,ld_iyh_c  ,ld_iyh_d   ,ld_iyh_e ,ld_iyh_iyh ,ld_iyh_iyl ,ld_h_xiy   ,ld_iyh_a,
  ld_iyl_b,ld_iyl_c  ,ld_iyl_d   ,ld_iyl_e ,ld_iyl_iyh ,ld_iyl_iyl ,ld_l_xiy   ,ld_iyl_a,
  ld_xiy_b,ld_xiy_c  ,ld_xiy_d   ,ld_xiy_e ,ld_xiy_h   ,ld_xiy_l   ,no_op      ,ld_xiy_a,
  no_op   ,no_op     ,no_op      ,no_op    ,ld_a_iyh   ,ld_a_iyl   ,ld_a_xiy   ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,add_a_iyh  ,add_a_iyl  ,add_a_xiy  ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,adc_a_iyh  ,adc_a_iyl  ,adc_a_xiy  ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,sub_iyh    ,sub_iyl    ,sub_xiy    ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,sbc_a_iyh  ,sbc_a_iyl  ,sbc_a_xiy  ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,and_iyh    ,and_iyl    ,and_xiy    ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,xor_iyh    ,xor_iyl    ,xor_xiy    ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,or_iyh     ,or_iyl     ,or_xiy     ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,cp_iyh     ,cp_iyl     ,cp_xiy     ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,fd_cb    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,pop_iy    ,no_op      ,ex_xsp_iy,no_op      ,push_iy    ,no_op      ,no_op   ,
  no_op   ,jp_iy     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,no_op     ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op   ,
  no_op   ,ld_sp_iy  ,no_op      ,no_op    ,no_op      ,no_op      ,no_op      ,no_op
};

Z80OPCODE cb(void) {

	byte	opcode;

	R.R++;
	opcode = M_RDMEM_OPCODE();
	Z80_ICount += cycles_cb[opcode];
	proc_cb(opcode);
}

Z80OPCODE dd(void) {

	byte	opcode;

	R.R++;
	opcode=M_RDMEM_OPCODE();
	Z80_ICount += cycles_xx[opcode];
	(*(opcode_dd[opcode]))();
}

Z80OPCODE ed(void) {

	byte	opcode;

	R.R++;
	opcode = M_RDMEM_OPCODE();
	Z80_ICount += cycles_ed[opcode];
	(*(opcode_ed[opcode]))();
}

Z80OPCODE fd(void) {

	byte	opcode;

	R.R++;
	opcode = M_RDMEM_OPCODE();
	Z80_ICount += cycles_xx[opcode];
	(*(opcode_fd[opcode]))();
}

static opcode_fn opcode_main[256]=
{
 nop     ,ld_bc_word,ld_xbc_a   ,inc_bc    ,inc_b   ,dec_b   ,ld_b_byte  ,rlca    ,
 ex_af_af,add_hl_bc ,ld_a_xbc   ,dec_bc    ,inc_c   ,dec_c   ,ld_c_byte  ,rrca    ,
 djnz    ,ld_de_word,ld_xde_a   ,inc_de    ,inc_d   ,dec_d   ,ld_d_byte  ,rla     ,
 jr      ,add_hl_de ,ld_a_xde   ,dec_de    ,inc_e   ,dec_e   ,ld_e_byte  ,rra     ,
 jr_nz   ,ld_hl_word,ld_xword_hl,inc_hl    ,inc_h   ,dec_h   ,ld_h_byte  ,daa     ,
 jr_z    ,add_hl_hl ,ld_hl_xword,dec_hl    ,inc_l   ,dec_l   ,ld_l_byte  ,cpl     ,
 jr_nc   ,ld_sp_word,ld_xbyte_a ,inc_sp    ,inc_xhl ,dec_xhl ,ld_xhl_byte,scf     ,
 jr_c    ,add_hl_sp ,ld_a_xbyte ,dec_sp    ,inc_a   ,dec_a   ,ld_a_byte  ,ccf     ,
 ld_b_b  ,ld_b_c    ,ld_b_d     ,ld_b_e    ,ld_b_h  ,ld_b_l  ,ld_b_xhl   ,ld_b_a  ,
 ld_c_b  ,ld_c_c    ,ld_c_d     ,ld_c_e    ,ld_c_h  ,ld_c_l  ,ld_c_xhl   ,ld_c_a  ,
 ld_d_b  ,ld_d_c    ,ld_d_d     ,ld_d_e    ,ld_d_h  ,ld_d_l  ,ld_d_xhl   ,ld_d_a  ,
 ld_e_b  ,ld_e_c    ,ld_e_d     ,ld_e_e    ,ld_e_h  ,ld_e_l  ,ld_e_xhl   ,ld_e_a  ,
 ld_h_b  ,ld_h_c    ,ld_h_d     ,ld_h_e    ,ld_h_h  ,ld_h_l  ,ld_h_xhl   ,ld_h_a  ,
 ld_l_b  ,ld_l_c    ,ld_l_d     ,ld_l_e    ,ld_l_h  ,ld_l_l  ,ld_l_xhl   ,ld_l_a  ,
 ld_xhl_b,ld_xhl_c  ,ld_xhl_d   ,ld_xhl_e  ,ld_xhl_h,ld_xhl_l,halt       ,ld_xhl_a,
 ld_a_b  ,ld_a_c    ,ld_a_d     ,ld_a_e    ,ld_a_h  ,ld_a_l  ,ld_a_xhl   ,ld_a_a  ,
 add_a_b ,add_a_c   ,add_a_d    ,add_a_e   ,add_a_h ,add_a_l ,add_a_xhl  ,add_a_a ,
 adc_a_b ,adc_a_c   ,adc_a_d    ,adc_a_e   ,adc_a_h ,adc_a_l ,adc_a_xhl  ,adc_a_a ,
 sub_b   ,sub_c     ,sub_d      ,sub_e     ,sub_h   ,sub_l   ,sub_xhl    ,sub_a   ,
 sbc_a_b ,sbc_a_c   ,sbc_a_d    ,sbc_a_e   ,sbc_a_h ,sbc_a_l ,sbc_a_xhl  ,sbc_a_a ,
 and_b   ,and_c     ,and_d      ,and_e     ,and_h   ,and_l   ,and_xhl    ,and_a   ,
 xor_b   ,xor_c     ,xor_d      ,xor_e     ,xor_h   ,xor_l   ,xor_xhl    ,xor_a   ,
 or_b    ,or_c      ,or_d       ,or_e      ,or_h    ,or_l    ,or_xhl     ,or_a    ,
 cp_b    ,cp_c      ,cp_d       ,cp_e      ,cp_h    ,cp_l    ,cp_xhl     ,cp_a    ,
 ret_nz  ,pop_bc    ,jp_nz      ,jp        ,call_nz ,push_bc ,add_a_byte ,rst_00  ,
 ret_z   ,ret       ,jp_z       ,cb        ,call_z  ,call    ,adc_a_byte ,rst_08  ,
 ret_nc  ,pop_de    ,jp_nc      ,out_byte_a,call_nc ,push_de ,sub_byte   ,rst_10  ,
 ret_c   ,exx       ,jp_c       ,in_a_byte ,call_c  ,dd      ,sbc_a_byte ,rst_18  ,
 ret_po  ,pop_hl    ,jp_po      ,ex_xsp_hl ,call_po ,push_hl ,and_byte   ,rst_20  ,
 ret_pe  ,jp_hl     ,jp_pe      ,ex_de_hl  ,call_pe ,ed      ,xor_byte   ,rst_28  ,
 ret_p   ,pop_af    ,jp_p       ,di        ,call_p  ,push_af ,or_byte    ,rst_30  ,
 ret_m   ,ld_sp_hl  ,jp_m       ,ei        ,call_m  ,fd      ,cp_byte    ,rst_38
};

Z80OPCODE ei(void) {

	byte	opcode;

/* If interrupts were disabled, execute one more instruction and check the */
/* IRQ line. If not, simply set interrupt flip-flop 2                      */

	if (!R.IFF1) {
#ifdef DEBUG
		if(DEBUG_FLG) {
			if(R.PC.D == Z80_Trap) {
				Z80_Trace=1;
			}
			if (Z80_Trace) {
				Z80_Debug(&R);
			}
		}
#endif
		R.IFF1 = R.IFF2 = 1;
		++R.R;
		opcode = M_RDMEM_OPCODE();
		Z80_ICount += cycles_main[opcode];
		(*(opcode_main[opcode]))();
		Interrupt();
	}
	else {
		R.IFF2 = 1;
	}
}

//**********************************************************************
//**********************************************************************

word Z80_GetPC(void) { return(R.PC.D); }
void Z80_SetPC(word adrs) { (R.PC.D)=adrs; } // KM 99'06/29

word Z80_IFF2(void){ return (R.IFF2); }; // KM 99'06/29
void Z80_GetRegs(Z80_Regs *Regs) { *Regs=R; }
void Z80_SetRegs(Z80_Regs *Regs) { R=*Regs; }

/****************************************************************************/
/* Set number of memory refresh wait states (i.e. extra cycles inserted     */
/* when the refresh register is being incremented)                          */
/****************************************************************************/

void Z80_SetWaitStates(int n) {

	int		i;

	for (i=0; i<256; ++i) {
		cycles_main[i] += n;
		cycles_cb[i] += n;
		cycles_ed[i] += n;
		cycles_xx[i] += n;
	}
}

void Z80_Cause_Interrupt(dword type) {

	if (type == Z80_NMI_INT) {
		R.pending_nmi = 1;
	}
	else if (type != Z80_IGNORE_INT) {
		R.pending_irq = type;
	}
}

void Z80_Clear_Pending_Interrupts(void) {

	R.pending_irq = Z80_IGNORE_INT;
	R.pending_nmi = 0;
}

/****************************************************************************/
/* Issue an interrupt if necessary                                          */
/****************************************************************************/

static void Interrupt(void) {

	if (R.pending_irq == Z80_IGNORE_INT && R.pending_nmi == 0) {
		return;			/* NS 970904 */
	}
	if (R.pending_nmi != 0 || R.IFF1) { /* NS 970904 */
		/* Clear interrupt flip-flop 1 */
		R.IFF1 = 0;
		/* Check if processor was halted */
		if (R.HALT) {
			++R.PC.W.l;
			R.HALT = 0;
		}
		/* if(j==Z80_NMI_INT) */
		if (R.pending_nmi != 0) { /* NS 970904 */
			R.pending_nmi = 0; /* NS 970904 */
			M_PUSH(R.PC.D);
			R.PC.D = 0x0066;
		}
		else {
			dword	j;
			j = R.pending_irq; /* NS 970904 */
			R.pending_irq = Z80_IGNORE_INT; /* NS 970904 */
			if (R.IM == 2) { /* Interrupt mode 2. Call [R.I:databyte] */
				M_PUSH(R.PC.D);
				R.PC.D = M_RDMEM_WORD(((word)j & 0xff) | ((int)R.I << 8));
			}
			else if (R.IM == 1) { /* Interrupt mode 1. RST 38h */
				Z80_ICount += cycles_main[0xFF];
				(*(opcode_main[0xFF]))();
			}
			else {
	/* Interrupt mode 0. We check for CALL and JP instructions, if neither  */
	/* of these were found we assume a 1 byte opcode was placed on the      */
	/* databus                                                              */
				switch(j & 0xFF0000l) {
					case 0xCD0000l: /* bugfix NS 970904 */
						M_PUSH(R.PC.D);
					case 0xC30000l: /* bugfix NS 970904 */
						R.PC.D = (word)(j & 0x0000FFFFl);
						break;
					default:
						j &= 0x000000ffl;
						Z80_ICount += cycles_main[(int)j];
						(*(opcode_main[(int)j]))();
					break;
				}
			}
		}
	}
}

/****************************************************************************/
/* Reset registers to their initial values                                  */
/****************************************************************************/
void Z80_Reset(void) {

static	byte	InitTables_virgin=1;
		int		i,p;
		byte	zs;

	Z80_Running = 1;
	memset(&R, 0, sizeof(Z80_Regs));
	R.SP.D = 0xF000;
	R.R = rand();
	Z80_Clear_Pending_Interrupts();

	if (InitTables_virgin) {
		InitTables_virgin = 0;
		for (i=0; i<256; ++i) {
			zs = 0;
			if (i == 0) {
				zs |= Z_FLAG;
			}
			if (i & 0x80) {
				zs |= S_FLAG;
			}
			p = 0;
			if (i &  1) ++p;
			if (i &  2) ++p;
			if (i &  4) ++p;
			if (i &  8) ++p;
			if (i & 16) ++p;
			if (i & 32) ++p;
			if (i & 64) ++p;
			if (i &128) ++p;
			PTable[i] = (byte)((p&1)? 0:V_FLAG);
			ZSTable[i] = zs;
			ZSPTable[i] = zs | PTable[i];
		}
		for (i=0; i<256; ++i) {
			ZSTable[i+256] = ZSTable[i] | (byte)C_FLAG;
			ZSPTable[i+256] = ZSPTable[i] | (byte)C_FLAG;
			PTable[i+256] = PTable[i] | (byte)C_FLAG;
		}
	}
	R.IFF1 = R.IFF2 = 1;								// 990302 puni
}

/****************************************************************************/
/* Execute IPeriod T-States. Return 0 if emulation should be stopped        */
/****************************************************************************/

word Z80_Execute(void) {

	byte	opcode;

	Z80_ICount = 0;
	if (R.pending_nmi != 0 || R.pending_irq != Z80_IGNORE_INT) {
		Interrupt();
	}
	R.R++;
	opcode = M_RDMEM_OPCODE();
	Z80_ICount += cycles_main[opcode];				// Thanx Mr.KM
	(*(opcode_main[opcode]))();
	return(Z80_ICount);
}

